#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of FormField represents a text area.
"""
import typing
import zlib
from decimal import Decimal

from borb.io.read.pdf_object import PDFObject
from borb.io.read.types import Boolean
from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, List, Name, Stream, String
from borb.pdf.canvas.layout.layout_element import Alignment
from borb.pdf.canvas.color.color import Color, HexColor, RGBColor
from borb.pdf.canvas.font.simple_font.font_type_1 import StandardType1Font
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.forms.form_field import FormField
from borb.pdf.page.page import Page


class TextArea(FormField):
    """
    This implementation of FormField represents a text area.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        background_color: typing.Optional[Color] = None,
        border_bottom: bool = True,
        border_color: Color = HexColor("808080"),
        border_left: bool = True,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = True,
        border_top: bool = True,
        border_width: Decimal = Decimal(1),
        default_value: str = "",
        field_name: typing.Optional[str] = None,
        font_color: Color = HexColor("000000"),
        font_size: typing.Optional[Decimal] = Decimal(12),
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: typing.Optional[Decimal] = Decimal(0),
        margin_left: typing.Optional[Decimal] = Decimal(0),
        margin_right: typing.Optional[Decimal] = Decimal(0),
        margin_top: typing.Optional[Decimal] = Decimal(0),
        number_of_lines: int = 5,
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        value: str = "",
        vertical_alignment: Alignment = Alignment.TOP,
    ):
        super(TextArea, self).__init__(
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            font_size=font_size,
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )
        assert number_of_lines > 0
        self._font_color = font_color
        self._value: str = value
        self._default_value: str = default_value
        self._number_of_lines: int = number_of_lines
        self._field_name: typing.Optional[str] = field_name
        self._widget_dictionary: typing.Optional[Dictionary] = None

    #
    # PRIVATE
    #

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        assert self._font_size is not None
        line_height: Decimal = self._font_size * Decimal(1.2)
        return Rectangle(
            available_space.get_x(),
            available_space.get_y()
            + available_space.get_height()
            - line_height * self._number_of_lines,
            max(available_space.get_width(), Decimal(64)),
            line_height * self._number_of_lines,
        )

    def _init_widget_dictionary(self, page: Page, layout_box: Rectangle) -> None:

        if self._widget_dictionary is not None:
            return

        root: typing.Optional[PDFObject] = page.get_root()
        assert root is not None
        assert isinstance(root, Dictionary)
        if "XRef" not in root:
            return

        # init page and font resources
        assert self._font_size is not None
        font_resource_name: Name = self._get_font_resource_name(
            StandardType1Font("Helvetica"), page
        )

        # widget resource dictionary
        widget_resources: Dictionary = Dictionary()
        widget_resources.set_is_unique(True)
        widget_resources[Name("Font")] = page["Resources"]["Font"]

        # widget normal appearance
        widget_normal_appearance: Stream = Stream()
        widget_normal_appearance.set_is_unique(True)  # type: ignore [attr-defined]
        widget_normal_appearance[Name("Type")] = Name("XObject")
        widget_normal_appearance[Name("Subtype")] = Name("Form")
        widget_normal_appearance[Name("BBox")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        widget_normal_appearance["BBox"].append(bDecimal(0))
        widget_normal_appearance["BBox"].append(bDecimal(0))
        widget_normal_appearance["BBox"].append(bDecimal(layout_box.width))
        widget_normal_appearance["BBox"].append(
            bDecimal((self._font_size + 1) * self._number_of_lines)
        )
        widget_normal_appearance[Name("Resources")] = widget_resources
        bts = b"/Tx BMC EMC"
        widget_normal_appearance[Name("DecodedBytes")] = bts
        widget_normal_appearance[Name("Bytes")] = zlib.compress(bts, 9)
        widget_normal_appearance[Name("Filter")] = Name("FlateDecode")
        widget_normal_appearance[Name("Length")] = bDecimal(len(bts))

        # widget appearance dictionary
        widget_appearance_dictionary: Dictionary = Dictionary()
        widget_appearance_dictionary.set_is_unique(True)
        widget_appearance_dictionary[Name("N")] = widget_normal_appearance

        # get Catalog
        catalog: Dictionary = root["XRef"]["Trailer"]["Root"]  # type: ignore [attr-defined]

        # widget dictionary
        # fmt: off
        self._widget_dictionary = Dictionary()
        self._widget_dictionary.set_is_unique(True)
        self._widget_dictionary[Name("Type")] = Name("Annot")
        self._widget_dictionary[Name("Subtype")] = Name("Widget")
        self._widget_dictionary[Name("F")] = bDecimal(4)
        self._widget_dictionary[Name("Rect")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        self._widget_dictionary["Rect"].append(bDecimal(layout_box.x))
        self._widget_dictionary["Rect"].append(bDecimal(layout_box.y + layout_box.height - (self._font_size + 1) * self._number_of_lines))
        self._widget_dictionary["Rect"].append(bDecimal(layout_box.x + layout_box.width))
        self._widget_dictionary["Rect"].append(bDecimal(layout_box.y + layout_box.height))
        self._widget_dictionary[Name("FT")] = Name("Tx")
        self._widget_dictionary[Name("Ff")] = bDecimal(4096)
        self._widget_dictionary[Name("P")] = catalog
        self._widget_dictionary[Name("T")] = String(self._field_name or self._get_auto_generated_field_name(page))
        self._widget_dictionary[Name("V")] = String(self._value)
        self._widget_dictionary[Name("DV")] = String(self._default_value)
        self._widget_dictionary[Name("DR")] = widget_resources
        # fmt: on

        # rendering instructions
        font_color_rgb: RGBColor = self._font_color.to_rgb()
        self._widget_dictionary[Name("DA")] = String(
            "%f %f %f rg /%s %f Tf"
            % (
                float(font_color_rgb.red),
                float(font_color_rgb.green),
                float(font_color_rgb.blue),
                font_resource_name,
                float(self._font_size),
            )
        )
        self._widget_dictionary[Name("AP")] = widget_appearance_dictionary

        # append field to page /Annots
        if "Annots" not in page:
            page[Name("Annots")] = List()
        page["Annots"].append(self._widget_dictionary)

        # append field to catalog
        if "AcroForm" not in catalog:
            catalog[Name("AcroForm")] = Dictionary()
            catalog["AcroForm"][Name("Fields")] = List()
            catalog["AcroForm"][Name("DR")] = widget_resources
            catalog["AcroForm"][Name("NeedAppearances")] = Boolean(True)
        catalog["AcroForm"]["Fields"].append(self._widget_dictionary)

    def _paint_content_box(self, page: "Page", available_space: Rectangle) -> None:

        # determine layout rectangle
        cbox: Rectangle = self._get_content_box(available_space)

        # init self._widget_dictionary
        self._init_widget_dictionary(page, cbox)

        # set location
        # fmt: off
        assert self._font_size is not None
        if self._widget_dictionary is not None:
            self._widget_dictionary["AP"]["N"]["BBox"][2] = bDecimal(cbox.get_width())
            self._widget_dictionary["AP"]["N"]["BBox"][3] = bDecimal(self._font_size)
            self._widget_dictionary["Rect"][0] = bDecimal(cbox.get_x())
            self._widget_dictionary["Rect"][1] = bDecimal(cbox.get_y())
            self._widget_dictionary["Rect"][2] = bDecimal(cbox.get_x() + cbox.get_width())
            self._widget_dictionary["Rect"][3] = bDecimal(cbox.get_y() + cbox.get_height())
        # fmt: on

    #
    # PUBLIC
    #



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of Event is triggered right after the Canvas has been processed.
"""
from borb.pdf.canvas.event.event_listener import Event
from borb.pdf.page.page import Page


class EndPageEvent(Event):
    """
    This implementation of Event is triggered right after the Canvas has been processed.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(self, page: Page):
        self._page: Page = page

    #
    # PRIVATE
    #

    #
    # PUBLIC
    #

    def get_page(self) -> Page:
        """
        This function returns the Page that triggered this BeginPageEvent
        """
        return self._page



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This class implements the basics for sending usage statistics to the borb server(s).
These kinds of checks are useful to get an idea of which functionality is most often used,
where development effort needs to be spent, etc
"""

import json
import sys
import threading
from datetime import datetime, timezone
import typing

import requests

from borb.license.license import License
from borb.license.machine_id import MachineID
from borb.license.persistent_random_user_id import PersistentRandomUserID
from borb.license.version import Version


class UsageStatistics:
    """
    This class implements the basics for sending usage statistics to the borb server(s).
    These kinds of checks are useful to get an idea of which functionality is most often used,
    where development effort needs to be spent, etc
    """

    # fmt: off
    _ENABLED: bool = True
    _ENDPOINT_URL: str = "https://cztmincfqq4fobtt6c7ect7gli0isbwx.lambda-url.us-east-1.on.aws/"
    # fmt: on

    #
    # CONSTRUCTOR
    #

    #
    # PRIVATE
    #

    @staticmethod
    def _get_machine_id() -> typing.Optional[str]:
        return MachineID.get()

    @staticmethod
    def _get_user_id() -> str:
        return License.get_user_id() or PersistentRandomUserID.get()

    @staticmethod
    def _send_usage_statistics_in_thread(event: str, document: typing.Optional["Document"] = None) -> None:  # type: ignore[name-defined]

        # get number_of_pages
        number_of_pages: int = 0
        try:
            if document is not None:
                number_of_pages = int(
                    document.get_document_info().get_number_of_pages()
                )
        except:
            pass

        # set payload
        json_payload: typing.Dict[str, typing.Any] = {
            "anonymous_user_id": UsageStatistics._get_user_id(),
            "company": License.get_company(),
            "event": event,
            "license_valid_from_in_ms": License.get_valid_from_in_ms(),
            "license_valid_until_in_ms": License.get_valid_until_in_ms(),
            "machine_id": UsageStatistics._get_machine_id(),
            "number_of_pages": number_of_pages,
            "sys_platform": sys.platform,
            "utc_time_in_ms": int(datetime.now(timezone.utc).timestamp() * 1000),
            "version": Version.get_version(),
        }

        # set headers
        headers = {"Content-type": "application/json", "Accept": "text/plain"}

        # perform request
        try:
            requests.post(
                UsageStatistics._ENDPOINT_URL,
                headers=headers,
                data=json.dumps(json_payload),
            )
        except:
            pass

    #
    # PUBLIC
    #

    @staticmethod
    def disable() -> None:
        """
        This function disables the sending of usage statistics
        :return:    None
        """
        UsageStatistics._ENABLED = False

    @staticmethod
    def enable() -> None:
        """
        This function enables the sending of usage statistics
        :return:    None
        """
        UsageStatistics._ENABLED = True

    @staticmethod
    def send_usage_statistics(event: str = "", document: typing.Optional["Document"] = None) -> None:  # type: ignore[name-defined]
        """
        This method sends the usage statistics to the borb license server
        :param event:       the event that is to be registered
        :param document     the Document being processed
        :return:        None
        """
        if not UsageStatistics._ENABLED:
            return
        try:
            threading.Thread(
                target=UsageStatistics._send_usage_statistics_in_thread,
                args=(
                    event,
                    document,
                ),
            ).start()
        except:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
TextRank – is a graph-based ranking model for text processing which can be used in order to find the most relevant sentences in text and also to find keywords.
The algorithm is explained in detail in the paper at https://web.eecs.umich.edu/~mihalcea/papers/mihalcea.emnlp04.pdf

In order to find relevant keywords, the textrank algorithm constructs a word network.
This network is constructed by looking which words follow one another.
A link is set up between two words if they follow one another, the link gets a higher weight if these 2 words occur more frequenctly next to each other in the text.

On top of the resulting network the Pagerank algorithm is applied to get the importance of each word.
The top 1/3 of all these words are kept and are considered relevant.
After this, a keywords table is constructed by combining the relevant words together if they appear following one another in the text.
"""
import io
import re
import typing

from borb.pdf.document.document import Document
from borb.pdf.canvas.canvas import Canvas
from borb.pdf.canvas.canvas_stream_processor import CanvasStreamProcessor
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.page.page import Page
from borb.toolkit.text.simple_text_extraction import SimpleTextExtraction
from borb.toolkit.text.stop_words import ENGLISH_STOP_WORDS


class TextRankKeywordExtraction(SimpleTextExtraction):
    """
    TextRank – is a graph-based ranking model for text processing which can be used in order to find the most relevant sentences in text and also to find keywords.
    The algorithm is explained in detail in the paper at https://web.eecs.umich.edu/~mihalcea/papers/mihalcea.emnlp04.pdf

    In order to find relevant keywords, the textrank algorithm constructs a word network.
    This network is constructed by looking which words follow one another.
    A link is set up between two words if they follow one another, the link gets a higher weight if these 2 words occur more frequenctly next to each other in the text.

    On top of the resulting network the Pagerank algorithm is applied to get the importance of each word.
    The top 1/3 of all these words are kept and are considered relevant.
    After this, a keywords table is constructed by combining the relevant words together if they appear following one another in the text.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(self):
        super().__init__()
        self._stopwords = [x.upper() for x in ENGLISH_STOP_WORDS]

        # keep track of keywords_per_page
        self._keywords_per_page: typing.Dict[
            int, typing.List[typing.Tuple[str, float]]
        ] = {}

    #
    # PRIVATE
    #

    def _end_page(self, page: Page):
        super()._end_page(page)

        # extract text
        txt: str = super(TextRankKeywordExtraction, self).get_text()[self._current_page]

        # transfer matrix
        mtx: typing.Dict[str, typing.Dict[str, float]] = {}

        # turn txt into lines
        lines = [x for x in re.split("\n*[.?!]+\n*", txt) if len(x) != 0]
        for line in lines:

            # split
            tokens: typing.List[str] = [
                x for x in re.split("[^A-Z]+", line.upper()) if len(x) > 3
            ]

            # build transfer matrix
            for i0 in range(0, len(tokens)):
                w0: str = tokens[i0].upper()
                if w0 not in mtx:
                    mtx[w0] = {}
                for i1 in range(-3, 3):
                    if i0 + i1 < 0 or i0 + i1 >= len(tokens) or i1 == 0:
                        continue
                    w1: str = tokens[i0 + i1].upper()
                    mtx[w0][w1] = mtx[w0].get(w1, 0) + 1

        # run eigenvalue algorithm
        ws: typing.List[str] = [x for x, _ in mtx.items()]
        eigenvalues_001: typing.Dict[str, float] = {x: 1 for x in ws}
        eigenvalues_002: typing.Dict[str, float] = {x: 0 for x in ws}
        delta: float = 1
        number_of_iterations: int = 0
        while delta > 0.0001 and number_of_iterations < 128:
            for w0 in ws:
                n: int = len(mtx[w0])
                for w1, f1 in mtx[w0].items():
                    eigenvalues_002[w1] += f1 * (eigenvalues_001[w0] / n)

            # calculate delta
            delta = max([abs(eigenvalues_001[x] - eigenvalues_002[x]) for x in ws])
            number_of_iterations += 1

            # update eigenvalues
            total_weight: float = sum([f for _, f in eigenvalues_002.items()])
            eigenvalues_001 = {
                x: (f / total_weight) for x, f in eigenvalues_002.items()
            }
            eigenvalues_002 = {x: 0 for x, _ in mtx.items()}

        # store keywords
        # fmt: off
        self._keywords_per_page[self._current_page] = [(x, f) for x, f in eigenvalues_001.items()]
        self._keywords_per_page[self._current_page].sort(key=lambda x: x[1], reverse=True)
        # fmt: on

    #
    # PUBLIC
    #

    def get_keywords(self) -> typing.Dict[int, typing.List[typing.Tuple[str, float]]]:
        """
        This function returns a typing.List[TextRankKeyword] for a given PDF
        """
        return self._keywords_per_page

    @staticmethod
    def get_keywords_from_pdf(
        pdf: Document,
    ) -> typing.Dict[int, typing.List[typing.Tuple[str, float]]]:
        """
        This function returns the keywords for a given PDF (per page)
        :param pdf:     the PDF to be analyzed
        :return:        the keywords per page (represented by typing.Dict[int, typing.List[typing.Tuple[str, float]]])
        """
        keywords_per_page: typing.Dict[int, typing.List[typing.Tuple[str, float]]] = {}
        number_of_pages: int = int(pdf.get_document_info().get_number_of_pages() or 0)
        for page_nr in range(0, number_of_pages):
            # get Page object
            page: Page = pdf.get_page(page_nr)
            page_source: io.BytesIO = io.BytesIO(page["Contents"]["DecodedBytes"])
            # register EventListener
            l: "TextRankKeywordExtraction" = TextRankKeywordExtraction()
            # process Page
            l._event_occurred(BeginPageEvent(page))
            CanvasStreamProcessor(page, Canvas(), []).read(page_source, [l])
            l._event_occurred(EndPageEvent(page))
            # add to output dictionary
            keywords_per_page[page_nr] = l.get_keywords()[0]
        # return
        return keywords_per_page



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This class is responsible for getting the machine GUID
"""
import subprocess
import sys
import typing


class MachineID:
    """
    This class is responsible for getting the machine GUID
    """

    #
    # CONSTRUCTOR
    #

    #
    # PRIVATE
    #

    @staticmethod
    def _run(cmd) -> typing.Optional[str]:
        try:
            return subprocess.run(
                cmd, shell=True, capture_output=True, check=True, encoding="utf-8"
            ).stdout.strip()
        except:
            return None

    #
    # PUBLIC
    #

    @staticmethod
    def get() -> typing.Optional[str]:
        """
        This function returns the machine UUID
        :return:    the machine UUID, or None
        """
        if sys.platform == "darwin":
            return MachineID._run(
                "ioreg -d2 -c IOPlatformExpertDevice | awk -F\\\" '/IOPlatformUUID/{print $(NF-1)}'",
            )
        if (
            sys.platform == "win32"
            or sys.platform == "cygwin"
            or sys.platform == "msys"
        ):
            return MachineID._run("wmic csproduct get uuid").split("\n")[2].strip()
        if sys.platform.startswith("linux"):
            return MachineID._run("cat /var/lib/dbus/machine-id") or MachineID._run(
                "cat /etc/machine-id"
            )
        if sys.platform.startswith("openbsd") or sys.platform.startswith("freebsd"):
            return MachineID._run("cat /etc/hostid") or MachineID._run(
                "kenv -q smbios.system.uuid"
            )
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement represents a heterogeneous Paragraph.
e.g. a Paragraph where one or more words are in bold (but not all of them)
"""
import copy
import typing
from decimal import Decimal

from borb.pdf.canvas.layout.image.image import Image
from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.emoji.emoji import Emoji
from borb.pdf.canvas.layout.layout_element import Alignment, LayoutElement
from borb.pdf.canvas.layout.text.chunk_of_text import ChunkOfText
from borb.pdf.canvas.layout.text.line_of_text import LineOfText
from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.page.page import Page


class LineBreakChunk(ChunkOfText):
    """
    This implementation of ChunkOfText represents a linebreak.
    This can be used in the HeterogeneousParagraph to force lines to be split
    """

    #
    # CONSTRUCTOR
    #

    def __init__(self):
        super(LineBreakChunk, self).__init__(text="\n")

    #
    # PRIVATE
    #

    #
    # PUBLIC
    #


class HeterogeneousParagraph(Paragraph):
    """
    This implementation of LayoutElement represents a heterogeneous collection of ChunkOfText elements.
    Furthermore (like Paragraph), this element has a default top- and bottom margin.
    e.g. a Paragraph where one or more words are in bold (but not all of them)
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        chunks_of_text: typing.List[
            typing.Union[ChunkOfText, LineOfText, Emoji, Image, str]
        ] = [],
        vertical_alignment: Alignment = Alignment.TOP,
        horizontal_alignment: Alignment = Alignment.LEFT,
        text_alignment: Alignment = Alignment.LEFT,
        border_top: bool = False,
        border_right: bool = False,
        border_bottom: bool = False,
        border_left: bool = False,
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_bottom_left: Decimal = Decimal(0),
        border_color: Color = HexColor("000000"),
        border_width: Decimal = Decimal(1),
        padding_top: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        margin_top: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        fixed_leading: typing.Optional[Decimal] = None,
        multiplied_leading: typing.Optional[Decimal] = None,
        background_color: typing.Optional[Color] = None,
    ):
        super(HeterogeneousParagraph, self).__init__(
            text="",
            text_alignment=text_alignment,
            vertical_alignment=vertical_alignment,
            horizontal_alignment=horizontal_alignment,
            border_top=border_top,
            border_right=border_right,
            border_bottom=border_bottom,
            border_left=border_left,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_bottom_left=border_radius_bottom_left,
            border_color=border_color,
            border_width=border_width,
            padding_top=padding_top,
            padding_right=padding_right,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            fixed_leading=fixed_leading,
            multiplied_leading=multiplied_leading,
            background_color=background_color,
        )
        # fmt: off
        self._chunks_of_text: typing.List[typing.Union[ChunkOfText, LineOfText, Emoji, Image, str]] = chunks_of_text
        # fmt: on

    #
    # PRIVATE
    #

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:

        # useful
        EMPTY_RECTANGLE: Rectangle = Rectangle(
            Decimal(0), Decimal(0), Decimal(0), Decimal(0)
        )

        # lines of text
        lines_of_text: typing.List[
            typing.List[LayoutElement]
        ] = self._split_to_lines_of_chunks_of_text(available_space)

        # determine height
        h: Decimal = Decimal(
            sum(
                [
                    max(
                        [
                            (
                                x.get_previous_layout_box() or EMPTY_RECTANGLE
                            ).get_height()
                            for x in line_of_text
                        ]
                    )
                    for line_of_text in lines_of_text
                ]
            )
        )

        # determine width
        w: Decimal = Decimal(
            max(
                [
                    sum(
                        [
                            (x.get_previous_layout_box() or EMPTY_RECTANGLE).get_width()
                            for x in line_of_text
                        ]
                    )
                    for line_of_text in lines_of_text
                ]
            )
        )
        if self._text_alignment == Alignment.JUSTIFIED:
            w = available_space.get_width()

        # return
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - h,
            w,
            h,
        )

    def _paint_content_box(self, page: Page, available_space: Rectangle) -> None:

        # lines of text
        lines_of_text: typing.List[
            typing.List[LayoutElement]
        ] = self._split_to_lines_of_chunks_of_text(available_space)

        for line in lines_of_text:
            for e in line:
                lbox: typing.Optional[Rectangle] = e.get_previous_layout_box()
                assert lbox is not None
                e.paint(page, lbox)

    def _split_to_lines_of_chunks_of_text(
        self, available_space: Rectangle
    ) -> typing.List[typing.List[LayoutElement]]:

        # useful
        EMPTY_RECTANGLE: Rectangle = Rectangle(
            Decimal(0), Decimal(0), Decimal(0), Decimal(0)
        )

        # build list
        initial_chunks_of_text: typing.List[LayoutElement] = []
        for c0 in self._chunks_of_text:
            if isinstance(c0, str):
                initial_chunks_of_text.append(
                    ChunkOfText(
                        c0,
                        border_top=self._border_top,
                        border_right=self._border_right,
                        border_bottom=self._border_bottom,
                        border_left=self._border_left,
                        border_radius_top_left=self._border_radius_top_left,
                        border_radius_top_right=self._border_radius_top_right,
                        border_radius_bottom_right=self._border_radius_bottom_right,
                        border_radius_bottom_left=self._border_radius_bottom_left,
                        background_color=self._background_color,
                        font_color=self._font_color,
                        font_size=self._font_size or Decimal(12),
                        font=self._font,
                    )
                )
                continue
            if isinstance(c0, Emoji):
                initial_chunks_of_text.append(c0)
                continue
            if isinstance(c0, Image):
                initial_chunks_of_text.append(c0)
                continue
            if isinstance(c0, ChunkOfText) and not isinstance(c0, LineOfText):
                initial_chunks_of_text.append(c0)
                continue
            if isinstance(c0, LineOfText):
                for word in c0.get_text().split(" "):
                    # TODO: continuation of borders when splitting LineOfText in ChunkOfText
                    initial_chunks_of_text.append(
                        ChunkOfText(
                            word + " ",
                            border_top=c0._border_top,
                            border_right=c0._border_right,
                            border_bottom=c0._border_bottom,
                            border_left=c0._border_left,
                            border_radius_top_left=c0._border_radius_top_left,
                            border_radius_top_right=c0._border_radius_top_right,
                            border_radius_bottom_left=c0._border_radius_bottom_left,
                            border_radius_bottom_right=c0._border_radius_bottom_right,
                            background_color=c0._background_color,
                            font_color=c0._font_color,
                            font_size=c0._font_size or Decimal(12),
                            font=c0._font,
                        )
                    )
                continue

        # perform initial layout (figure out where to break lines)
        next_x: Decimal = available_space.get_x()
        line_width: Decimal = Decimal(0)
        line: typing.List[LayoutElement] = []
        lines: typing.List[typing.List[LayoutElement]] = []
        for c1 in initial_chunks_of_text:
            w: Decimal = c1.get_layout_box(
                Rectangle(
                    next_x,
                    available_space.get_y(),
                    available_space.get_width() - line_width,
                    available_space.get_height(),
                )
            ).get_width()
            if (line_width + w > available_space.get_width()) or (
                isinstance(c1, LineBreakChunk)
            ):
                # get ready for next line
                lines.append(copy.deepcopy(line))
                next_x = available_space.get_x() + w
                line_width = w
                # fix the one element that does not fit
                assert c1._previous_layout_box is not None
                c1._previous_layout_box.x = available_space.get_x()
                line = [c1]
                continue
            else:
                next_x += w
                line_width += w
                line.append(c1)
        if len(line) > 0:
            lines.append(copy.deepcopy(line))

        # update ys
        e_prev_layout_box: Rectangle = EMPTY_RECTANGLE
        prev_y: Decimal = available_space.get_y() + available_space.get_height()
        for line in lines:
            line_height: Decimal = max(
                [
                    (x.get_previous_layout_box() or EMPTY_RECTANGLE).get_height()
                    for x in line
                ]
            )
            y: Decimal = prev_y - line_height
            for c1 in line:
                # the next line makes the type-checker happy
                e_prev_layout_box = c1.get_previous_layout_box() or EMPTY_RECTANGLE
                delta_height: Decimal = line_height - e_prev_layout_box.get_height()
                c1.get_layout_box(
                    Rectangle(
                        e_prev_layout_box.get_x(),
                        y,
                        e_prev_layout_box.get_width(),
                        line_height - delta_height,
                    )
                )
            prev_y -= line_height

        # update xs (text_alignment == CENTERED)
        delta_x: Decimal = Decimal(0)
        line_max_x: Decimal = Decimal(0)
        cbox_max_x: Decimal = Decimal(0)
        if self._text_alignment == Alignment.CENTERED:
            for line in lines:
                line_max_x = (
                    line[-1].get_previous_layout_box() or EMPTY_RECTANGLE
                ).get_x() + (
                    line[-1].get_previous_layout_box() or EMPTY_RECTANGLE
                ).get_width()
                cbox_max_x = available_space.get_x() + available_space.get_width()
                delta_x = (cbox_max_x - line_max_x) / 2
                for c1 in line:
                    assert c1._previous_layout_box is not None
                    c1._previous_layout_box.x += delta_x

        # update xs (text_alignment == LEFT)
        if self._text_alignment == Alignment.RIGHT:
            for line in lines:
                line_max_x = (
                    line[-1].get_previous_layout_box() or EMPTY_RECTANGLE
                ).get_x() + (
                    line[-1].get_previous_layout_box() or EMPTY_RECTANGLE
                ).get_width()
                cbox_max_x = available_space.get_x() + available_space.get_width()
                delta_x = cbox_max_x - line_max_x
                for c1 in line:
                    assert c1._previous_layout_box is not None
                    c1._previous_layout_box.x += delta_x

        # update xs (text_alignment == JUSTIFIED)
        if self._text_alignment == Alignment.JUSTIFIED:
            for line in lines:
                line_max_x = (
                    line[-1].get_previous_layout_box() or EMPTY_RECTANGLE
                ).get_x() + (
                    line[-1].get_previous_layout_box() or EMPTY_RECTANGLE
                ).get_width()
                cbox_max_x = available_space.get_x() + available_space.get_width()
                delta_x = (cbox_max_x - line_max_x) / (len(line) - 1)
                for c1 in line:
                    assert c1._previous_layout_box is not None
                    c1._previous_layout_box.x += delta_x

        # return
        return lines

    #
    # PUBLIC
    #



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A universally unique identifier (UUID) is a 128-bit label used for information in computer systems.
The term globally unique identifier (GUID) is also used.
When generated according to the standard methods, UUIDs are, for practical purposes, unique.
Their uniqueness does not depend on a central registration authority or coordination between the parties generating them,
unlike most other numbering schemes.
While the probability that a UUID will be duplicated is not zero, it is close enough to zero to be negligible.
"""

import random


class UUID:
    """
    A universally unique identifier (UUID) is a 128-bit label used for information in computer systems.
    The term globally unique identifier (GUID) is also used.
    When generated according to the standard methods, UUIDs are, for practical purposes, unique.
    Their uniqueness does not depend on a central registration authority or coordination between the parties generating them,
    unlike most other numbering schemes.
    While the probability that a UUID will be duplicated is not zero, it is close enough to zero to be negligible.
    """

    #
    # CONSTRUCTOR
    #

    #
    # PRIVATE
    #

    #
    # PUBLIC
    #

    @staticmethod
    def get() -> str:
        """
        This function returns a randomly generated UUID
        :return:    a randomly generated UUID
        """
        return "".join(
            [
                "".join([random.choice("0123456789abcdef") for _ in range(0, l)]) + "-"
                for l in [8, 4, 4, 4, 12]
            ]
        )[:-1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class represents a simple Markov model to generate text.
"""
import json
import random
import re
import typing
import zlib
from pathlib import Path

import requests


class TextGenerator:
    """
    This class represents a simple Markov model to generate text.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(self):
        self._token_ids: typing.Dict[str, int] = {}
        self._rev_token_ids: typing.Dict[int, str] = {}
        self._token_frequency: typing.Dict[int, int] = {}
        self._markov_model: typing.Dict[
            typing.Tuple[int, int], typing.Dict[int, int]
        ] = {}

    #
    # PRIVATE
    #

    def __len__(self):
        return len(self._token_ids)

    @staticmethod
    def _is_valid_token(s: str) -> bool:
        # forbid tokens of length 0
        if len(s) == 0:
            return False
        # forbid tokens that contain non-alphanumeric characters
        if any([not (x.isalnum() or x in ".?!,") for x in s]):
            return False
        # forbid tokens that consist solely of uppercase
        if s.upper() == s:
            return False
        # forbid tokens that consist solely of numbers
        if s.isnumeric():
            return False
        return True

    def _train_using_project_gutenberg(self, gutenberg_url: str) -> "TextGenerator":

        training_text: str = requests.get(gutenberg_url).text

        # trim header:
        header_text: str = "*** START OF THE PROJECT GUTENBERG EBOOK"
        if header_text in training_text:
            start_pos: int = training_text.find(
                "***", training_text.find(header_text) + len(header_text)
            ) + len("***")
            training_text = training_text[start_pos:]
        # trim footer
        footer_text: str = "*** END OF THE PROJECT GUTENBERG EBOOK"
        if footer_text in training_text:
            end_pos: int = training_text.find(footer_text)
            training_text = training_text[0:end_pos]
        # tokenize, build model
        tokens: typing.List[str] = re.split("[ \n\t\r]+", training_text)
        self._token_ids = {}
        self._rev_token_ids = {}
        self._token_frequency = {}
        self._markov_model = {}
        for i in range(0, len(tokens) - 2):

            # get token, set/create ID
            t0: str = tokens[i].strip(" \n\t")
            if not TextGenerator._is_valid_token(t0):
                continue
            t0_id: int = self._token_ids.get(t0, len(self._token_ids))
            self._token_ids[t0] = t0_id

            # get token, set/create ID
            t1: str = tokens[i + 1].strip(" \n\t")
            if not TextGenerator._is_valid_token(t1):
                continue
            t1_id: int = self._token_ids.get(t1, len(self._token_ids))
            self._token_ids[t1] = t1_id

            # get token, set/create ID
            t2: str = tokens[i + 2].strip(" \n\t")
            if not TextGenerator._is_valid_token(t2):
                continue
            t2_id: int = self._token_ids.get(t2, len(self._token_ids))
            self._token_ids[t2] = t2_id

            # update frequency
            self._token_frequency[t0_id] = self._token_frequency.get(t0_id, 0) + 1

            # update markov model
            mm_key: typing.Tuple[int, int] = (t0_id, t1_id)
            if mm_key not in self._markov_model:
                self._markov_model[mm_key] = {}
            self._markov_model[mm_key][t2_id] = (
                self._markov_model[mm_key].get(t2_id, 0) + 1
            )
        # reverse ID map
        self._rev_token_ids = {v: k for k, v in self._token_ids.items()}

        # return
        return self

    #
    # PUBLIC
    #

    def generate(self, min_sentence_length: int = 32) -> str:
        """
        This function generates a string using the statistical model
        stored in this TextGenerator
        :param min_sentence_length:     the minimal sentence length (in tokens)
        :return:                        a string
        """
        # build some text
        sentence_being_built: typing.List[str] = []

        # build text
        while len(
            sentence_being_built
        ) < min_sentence_length or not sentence_being_built[-1].endswith("."):

            # init
            if len(sentence_being_built) == 0:
                seed: typing.Tuple[int, int] = random.choice(
                    [
                        (t0, t1)
                        for (t0, t1) in self._markov_model.keys()
                        if len(self._rev_token_ids[t0]) > 0
                        and self._rev_token_ids[t0][0]
                        == self._rev_token_ids[t0][0].upper()
                    ]
                )
                sentence_being_built.append(self._rev_token_ids[seed[0]])
                sentence_being_built.append(self._rev_token_ids[seed[1]])
                continue

            # default
            t0_id: int = self._token_ids[sentence_being_built[-2]]
            t1_id: int = self._token_ids[sentence_being_built[-1]]

            # build key
            mm_key: typing.Tuple[int, int] = (t0_id, t1_id)
            if mm_key not in self._markov_model:
                sentence_being_built = []
                continue

            # select next element
            nexts: typing.Dict[int, int] = self._markov_model[mm_key]
            ops: typing.List[int] = []
            for k, v in nexts.items():
                for _ in range(0, v):
                    ops.append(k)
            if len(sentence_being_built) < min_sentence_length:
                sentence_being_built.append(self._rev_token_ids[random.choice(ops)])
            else:
                end_sentence_ops: typing.List[int] = [
                    x for x in ops if self._rev_token_ids[x].endswith(".")
                ]
                if len(end_sentence_ops) == 0:
                    sentence_being_built.append(self._rev_token_ids[random.choice(ops)])
                else:
                    sentence_being_built.append(
                        self._rev_token_ids[random.choice(end_sentence_ops)]
                    )

        # return
        return "".join([x + " " for x in sentence_being_built])

    def load(self, json_file: Path) -> "TextGenerator":
        """
        This function loads a TextGenerator from a (zipped) JSON file
        :param json_file:   the location where to load the (zipped) JSON
        :return:            self
        """
        with open(json_file, "rb") as json_file_handle:
            json_obj = json.loads(zlib.decompress(json_file_handle.read()))
        (
            self._token_ids,
            self._rev_token_ids,
            self._token_frequency,
            self._markov_model,
        ) = (
            {k: int(v) for k, v in json_obj["token_ids"].items()},
            {int(k): v for k, v in json_obj["rev_token_ids"].items()},
            {int(k): int(v) for k, v in json_obj["token_frequency"].items()},
            {
                (int(k.split("|")[0]), int(k.split("|")[1])): {
                    int(a): int(b) for a, b in v.items()
                }
                for k, v in json_obj["markov_model"].items()
            },
        )
        return self

    def store(self, file_name: str) -> None:
        """
        This function stores this TextGenerator in a (zipped) JSON format
        :param file_name:   the location where to store the (zipped) JSON
        :return:            None
        """
        with open(file_name, "wb") as json_file_handle:
            json_file_handle.write(
                zlib.compress(
                    bytes(
                        json.dumps(
                            {
                                "token_ids": self._token_ids,
                                "rev_token_ids": self._rev_token_ids,
                                "token_frequency": self._token_frequency,
                                "markov_model": {
                                    str(k[0]) + "|" + str(k[1]): v
                                    for k, v in self._markov_model.items()
                                },
                            }
                        ),
                        encoding="utf8",
                    ),
                    level=9,
                )
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This class is responsible for getting/setting an anonymous (GDPR-compliant) ID
"""
import sysconfig
import typing
from pathlib import Path

from borb.license.uuid import UUID


class PersistentRandomUserID:
    """
    This class is responsible for getting/setting an anonymous (GDPR-compliant) ID
    """

    USER_ID_FILE_NAME: str = "anonymous_user_id"
    USER_ID: typing.Optional[str] = None

    #
    # CONSTRUCTOR
    #

    #
    # PRIVATE
    #

    @staticmethod
    def _get_borb_installation_dir() -> typing.Optional[Path]:
        for path_name in sysconfig.get_path_names():
            # check whether the installation path directory exists
            installation_path: Path = Path(sysconfig.get_path(path_name))
            if not installation_path.exists():
                continue
            # check whether the borb directory exists
            borb_dir: Path = installation_path / "borb"
            if borb_dir.exists():
                return borb_dir
        return None

    @staticmethod
    def _get_user_id_file_from_borb_dir() -> typing.Optional[Path]:
        installation_dir: typing.Optional[
            Path
        ] = PersistentRandomUserID._get_borb_installation_dir()
        if installation_dir is None:
            return None
        # check whether the USER_ID_FILE_NAME file exists
        user_id_file: Path = installation_dir / PersistentRandomUserID.USER_ID_FILE_NAME
        if user_id_file.exists():
            return user_id_file
        # return
        return None

    #
    # PUBLIC
    #

    @staticmethod
    def get() -> typing.Optional[str]:
        """
        This function (creates and then) returns an anonymous user ID.
        This ID is stored in a file in the borb installation directory to ensure consistency between calls.
        :return:    an anonymous user ID
        """
        # IF borb installation directory exists, but the file does not exist (yet)
        # THEN create the file, return the uuid
        installation_dir: typing.Optional[
            Path
        ] = PersistentRandomUserID._get_borb_installation_dir()
        user_id_file: typing.Optional[
            Path
        ] = PersistentRandomUserID._get_user_id_file_from_borb_dir()
        if (
            installation_dir is not None
            and installation_dir.exists()
            and (user_id_file is None or not user_id_file.exists())
        ):
            try:
                # fmt: off
                new_uuid: str = UUID.get()
                with open(installation_dir / PersistentRandomUserID.USER_ID_FILE_NAME, "w") as fh:
                    fh.write(new_uuid)
                return new_uuid
                # fmt: on
            except:
                pass

        # IF the borb installation directory exists, and the user_id file exists
        # THEN read the user_id file, and return its content
        if user_id_file is not None and user_id_file.exists():
            prev_uuid: typing.Optional[str] = None
            try:
                with open(user_id_file, "r") as fh:
                    prev_uuid = fh.read()
            except:
                pass
            return prev_uuid

        # default
        return None



"""
    This file is part of the borb (TM) project.
    Copyright (c) 2020-2040 borb (EZ)
    Authors: Joris Schellekens, et al.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3
    as published by the Free Software Foundation with the addition of the
    following permission added to Section 15 as permitted in Section 7(a):
    FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY
    BORB GROUP. BORB GROUP DISCLAIMS THE WARRANTY OF NON INFRINGEMENT
    OF THIRD PARTY RIGHTS

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.

    See the GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program; if not, see http://www.gnu.org/licenses or write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA, 02110-1301 USA.

    The interactive user interfaces in modified source and object code versions
    of this program must display Appropriate Legal Notices, as required under
    Section 5 of the GNU Affero General Public License.
    In accordance with Section 7(b) of the GNU Affero General Public License,
    a covered work must retain the producer line in every PDF that is created
    or manipulated using borb.

    You can be released from the requirements of the license by purchasing
    a commercial license. Buying such a license is mandatory as soon as you
    develop commercial activities involving the borb software without
    disclosing the source code of your own applications.

    These activities include: offering paid services to customers as an ASP,
    serving PDFs on the fly in a web application, shipping borb with a closed
    source product.

    For more information, please contact borb Software Corp. at this
    address: joris.schellekens.1989@gmail.com
"""

# Color
from .canvas.color.color import (
    Color,
    RGBColor,
    CMYKColor,
    GrayColor,
    HSVColor,
    HexColor,
    HSVColor,
    X11Color,
)
from .canvas.color.pantone import Pantone

# Equation
from .canvas.layout.equation.equation import Equation

# Image
from .canvas.layout.image.barcode import Barcode
from .canvas.layout.image.barcode import BarcodeType
from .canvas.layout.image.chart import Chart
from .canvas.layout.image.image import Image
from .canvas.layout.layout_element import Alignment

# List
from .canvas.layout.list.list import List
from .canvas.layout.list.ordered_list import OrderedList
from .canvas.layout.list.roman_list import RomanNumeralOrderedList
from .canvas.layout.list.unordered_list import UnorderedList

# PageLayout
from .canvas.layout.page_layout.page_layout import PageLayout
from .canvas.layout.page_layout.multi_column_layout import MultiColumnLayout
from .canvas.layout.page_layout.multi_column_layout import SingleColumnLayout
from .canvas.layout.page_layout.single_column_layout_with_overflow import (
    SingleColumnLayoutWithOverflow,
)

# Flow
from .canvas.layout.page_layout.inline_flow import InlineFlow
from .canvas.layout.page_layout.block_flow import BlockFlow

# Shape
from .canvas.line_art.line_art_factory import LineArtFactory
from .canvas.layout.shape.connected_shape import ConnectedShape
from .canvas.layout.shape.disconnected_shape import DisconnectedShape
from .canvas.layout.smart_art.smart_art import SmartArt

# Table
from .canvas.layout.table.fixed_column_width_table import FixedColumnWidthTable
from .canvas.layout.table.flexible_column_width_table import FlexibleColumnWidthTable
from .canvas.layout.table.table import Table, TableCell
from .canvas.layout.table.table_util import TableUtil
from .canvas.layout.text.heading import Heading

# Forms
from .canvas.layout.forms.form_field import FormField
from .canvas.layout.forms.text_field import TextField
from .canvas.layout.forms.text_area import TextArea
from .canvas.layout.forms.drop_down_list import DropDownList
from .canvas.layout.forms.country_drop_down_list import CountryDropDownList
from .canvas.layout.forms.check_box import CheckBox
from .canvas.layout.forms.push_button import PushButton, JavaScriptPushButton

# Paragraph
from .canvas.layout.text.paragraph import Paragraph
from .canvas.layout.text.heterogeneous_paragraph import HeterogeneousParagraph
from .canvas.layout.text.heading import Heading
from .canvas.layout.text.chunk_of_text import ChunkOfText
from .canvas.lipsum.lipsum import Lipsum

# Document, Page, PDF
from .document.document import Document
from .page.page import Page
from .pdf import PDF



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener extracts all text from a PDF Document,
    substituting composite glyphs for their simpler representations
"""
import io
import typing

from borb.pdf.document.document import Document
from borb.pdf.page.page import Page
from borb.pdf.canvas.canvas import Canvas
from borb.pdf.canvas.canvas_stream_processor import CanvasStreamProcessor
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.toolkit.text.simple_text_extraction import SimpleTextExtraction


class SimpleNonLigatureTextExtraction(SimpleTextExtraction):
    """
    This implementation of EventListener extracts all text from a PDF Document,
    substituting composite glyphs for their simpler representations
    """

    #
    # CONSTRUCTOR
    #

    def __init__(self):
        super(SimpleNonLigatureTextExtraction, self).__init__()
        self._ligatures_to_replace: typing.Dict[str, str] = {
            "Ꜳ": "AA",
            "ꜳ": "aa",
            "Æ": "AE",
            "æ": "ae",
            "ꬱ": "aə",
            "Ꜵ": "AO",
            "ꜵ": "ao",
            "Ꜷ": "AU",
            "ꜷ": "au",
            "Ꜹ": "AV",
            "ꜹ": "av",
            "Ꜻ": "AV",
            "ꜻ": "av",
            "Ꜽ": "AY",
            "ꜽ": "ay",
            "🙰": "et",
            "ꭁ": "əø",
            "ﬀ": "ff",
            "ﬃ": "ffi",
            "ﬄ": "ffl",
            "ﬁ": "fi",
            "ﬂ": "fl",
            "℔": "lb",
            "Ỻ": "IL",
            "ỻ": "ll",
            "Œ": "OE",
            "œ": "oe",
            "Ꝏ": "OO",
            "ꝏ": "oo",
            "ꭢ": "ɔe",
            "ſs": "ẞ",
            "ſz": "ß",
            "ﬆ": "st",
            "ﬅ": "ſt",
            "Ꜩ": "TZ",
            "ꜩ": "tz",
            "ᵫ": "ue",
            "ꭣ": "uo",
            "W": "VV",
            "w": "vv",
            "Ꝡ": "VY",
            "ꝡ": "vy",
        }

    #
    # PRIVATE
    #

    #
    # PUBLIC
    #

    def get_text(self) -> typing.Dict[int, str]:
        """
        This function returns all text on a given PDF
        """
        out: typing.Dict[int, str] = {}
        for k0, v0 in self._text_per_page.items():
            text = self._text_per_page[k0]
            while any([k1 in text for k1, v1 in self._ligatures_to_replace.items()]):
                for k3, v3 in self._ligatures_to_replace.items():
                    text = text.replace(k3, v3)
            out[k0] = text
        return out

    @staticmethod
    def get_text_from_pdf(pdf: Document) -> typing.Dict[int, str]:
        """
        This function returns the text for a given PDF (per page)
        :param pdf:     the PDF to be analyzed
        :return:        the text per page (represented by typing.Dict[int, str])
        """
        text_per_page: typing.Dict[int, str] = {}
        number_of_pages: int = int(pdf.get_document_info().get_number_of_pages() or 0)
        for page_nr in range(0, number_of_pages):
            # get Page object
            page: Page = pdf.get_page(page_nr)
            page_source: io.BytesIO = io.BytesIO(page["Contents"]["DecodedBytes"])
            # register EventListener
            l: "SimpleNonLigatureTextExtraction" = SimpleNonLigatureTextExtraction()
            # process Page
            l._event_occurred(BeginPageEvent(page))
            CanvasStreamProcessor(page, Canvas(), []).read(page_source, [l])
            l._event_occurred(EndPageEvent(page))
            # add to output dictionary
            text_per_page[page_nr] = l.get_text()[0]
        # return
        return text_per_page



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This class contains the basics for getting/setting version and license information
within the borb project
"""


class Version:
    """
    This class contains the basics for getting/setting version and license information
    within the borb project
    """

    #
    # CONSTRUCTOR
    #

    #
    # PRIVATE
    #

    #
    # PUBLIC
    #

    @staticmethod
    def get_author() -> str:
        """
        This function returns the author of the borb library
        :return:    the author of the borb library
        """
        return "Joris Schellekens"

    @staticmethod
    def get_producer() -> str:
        """
        This function returns the producer line added to each PDF made by borb
        :return:    the PDF producer
        """
        return "borb"

    @staticmethod
    def get_version() -> str:
        """
        This function returns the current borb version
        :return:    the current borb version
        """
        return "2.1.12"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Lorem Ipsum is simply dummy text of the printing and typesetting industry.
    Lorem Ipsum has been the industry's standard dummy text ever since the 1500s,
    when an unknown printer took a galley of type and scrambled it to make a type specimen book.
    It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged.
    It was popularised in the 1960s with the release of Letraset sheets containing Lorem Ipsum passages,
    and more recently with desktop publishing software like Aldus PageMaker including versions of Lorem Ipsum.
"""

import random
from pathlib import Path

from borb.pdf.canvas.lipsum.text_generator import TextGenerator


class Lipsum:
    """
    Lorem Ipsum is simply dummy text of the printing and typesetting industry.
    Lorem Ipsum has been the industry's standard dummy text ever since the 1500s,
    when an unknown printer took a galley of type and scrambled it to make a type specimen book.
    It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged.
    It was popularised in the 1960s with the release of Letraset sheets containing Lorem Ipsum passages,
    and more recently with desktop publishing software like Aldus PageMaker including versions of Lorem Ipsum.
    """

    #
    # CONSTRUCTOR
    #

    #
    # PRIVATE
    #

    #
    # PUBLIC
    #

    @staticmethod
    def generate_agatha_christie_text(number_of_sentences: int = 5) -> str:
        """
        This function produces Agatha Christie styled lorem ipsum text
        :param number_of_sentences:         the number of sentences to be produced
        :return:                            lorem ipsum text
        """
        assert (
            number_of_sentences >= 1
        ), "number_of_sentences must be a positive non-zero quantity"
        resources_dir: Path = Path(__file__).parent / "resources"
        tg: TextGenerator = TextGenerator().load(
            resources_dir / "mm_agatha_christie.json"
        )
        return "".join(
            [
                tg.generate(random.randint(7, 32)) + " "
                for _ in range(0, number_of_sentences)
            ]
        )

    @staticmethod
    def generate_alan_alexander_milne_text(number_of_sentences: int = 5) -> str:
        """
        This function produces A.A. Milne styled lorem ipsum text
        :param number_of_sentences:         the number of sentences to be produced
        :return:                            lorem ipsum text
        """
        assert (
            number_of_sentences >= 1
        ), "number_of_sentences must be a positive non-zero quantity"
        resources_dir: Path = Path(__file__).parent / "resources"
        tg: TextGenerator = TextGenerator().load(
            resources_dir / "mm_alan_alexander_milne.json"
        )
        return "".join(
            [
                tg.generate(random.randint(7, 32)) + " "
                for _ in range(0, number_of_sentences)
            ]
        )

    @staticmethod
    def generate_arthur_conan_doyle_text(number_of_sentences: int = 5) -> str:
        """
        This function produces Arthur Conan Doyle styled lorem ipsum text
        :param number_of_sentences:         the number of sentences to be produced
        :return:                            lorem ipsum text
        """
        assert (
            number_of_sentences >= 1
        ), "number_of_sentences must be a positive non-zero quantity"
        resources_dir: Path = Path(__file__).parent / "resources"
        tg: TextGenerator = TextGenerator().load(
            resources_dir / "mm_arthur_conan_doyle.json"
        )
        return "".join(
            [
                tg.generate(random.randint(7, 32)) + " "
                for _ in range(0, number_of_sentences)
            ]
        )

    @staticmethod
    def generate_emily_bronte_text(number_of_sentences: int = 5) -> str:
        """
        This function produces Emily Bronte styled lorem ipsum text
        :param number_of_sentences:         the number of sentences to be produced
        :return:                            lorem ipsum text
        """
        assert (
            number_of_sentences >= 1
        ), "number_of_sentences must be a positive non-zero quantity"
        resources_dir: Path = Path(__file__).parent / "resources"
        tg: TextGenerator = TextGenerator().load(resources_dir / "mm_emily_bronte.json")
        return "".join(
            [
                tg.generate(random.randint(7, 32)) + " "
                for _ in range(0, number_of_sentences)
            ]
        )

    @staticmethod
    def generate_jane_austen_text(number_of_sentences: int = 5) -> str:
        """
        This function produces Jane Austen styled lorem ipsum text
        :param number_of_sentences:         the number of sentences to be produced
        :return:                            lorem ipsum text
        """
        assert (
            number_of_sentences >= 1
        ), "number_of_sentences must be a positive non-zero quantity"
        resources_dir: Path = Path(__file__).parent / "resources"
        tg: TextGenerator = TextGenerator().load(resources_dir / "mm_jane_austen.json")
        return "".join(
            [
                tg.generate(random.randint(7, 32)) + " "
                for _ in range(0, number_of_sentences)
            ]
        )

    @staticmethod
    def generate_lewis_carroll_text(number_of_sentences: int = 5) -> str:
        """
        This function produces Lewis Carroll styled lorem ipsum text
        :param number_of_sentences:         the number of sentences to be produced
        :return:                            lorem ipsum text
        """
        assert (
            number_of_sentences >= 1
        ), "number_of_sentences must be a positive non-zero quantity"
        resources_dir: Path = Path(__file__).parent / "resources"
        tg: TextGenerator = TextGenerator().load(
            resources_dir / "mm_lewis_carroll.json"
        )
        return "".join(
            [
                tg.generate(random.randint(7, 32)) + " "
                for _ in range(0, number_of_sentences)
            ]
        )

    @staticmethod
    def generate_lipsum_text(number_of_sentences: int = 5) -> str:
        """
        This function produces lorem ipsum text
        :param number_of_sentences:         the number of sentences to be produced
        :param start_with_lorem_ipsum:      whether or not to start the text with "Lorem ipsum"
        :return:                            lorem ipsum text
        """
        assert (
            number_of_sentences >= 1
        ), "number_of_sentences must be a positive non-zero quantity"
        resources_dir: Path = Path(__file__).parent / "resources"
        tg: TextGenerator = TextGenerator().load(resources_dir / "mm_lipsum.json")
        return "".join(
            [
                tg.generate(random.randint(7, 32)) + " "
                for _ in range(0, number_of_sentences)
            ]
        )

    @staticmethod
    def generate_mary_shelley_text(number_of_sentences: int = 5) -> str:
        """
        This function produces Mary Shelley styled lorem ipsum text
        :param number_of_sentences:         the number of sentences to be produced
        :return:                            lorem ipsum text
        """
        assert (
            number_of_sentences >= 1
        ), "number_of_sentences must be a positive non-zero quantity"
        resources_dir: Path = Path(__file__).parent / "resources"
        tg: TextGenerator = TextGenerator().load(resources_dir / "mm_mary_shelley.json")
        return "".join(
            [
                tg.generate(random.randint(7, 32)) + " "
                for _ in range(0, number_of_sentences)
            ]
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of EventListener exports a Page as an mp3 file, essentially reading the text on the Page
"""
import io
import tempfile
import typing
from decimal import Decimal
from pathlib import Path

from gtts import gTTS  # type: ignore [import]

from borb.pdf.document.document import Document
from borb.pdf.canvas.canvas import Canvas
from borb.pdf.canvas.canvas_stream_processor import CanvasStreamProcessor
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.page.page import Page
from borb.toolkit.text.simple_paragraph_extraction import SimpleParagraphExtraction


class PDFToMP3(SimpleParagraphExtraction):
    """
    This implementation of EventListener exports a Page as an mp3 file (returned as bytes), essentially reading the text on the Page
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        include_position: bool = True,
        language: str = "en",
        slow: bool = False,
    ):
        """
        Constructs a new PDFToMP3
        """
        super(PDFToMP3, self).__init__()

        # tts info
        self._include_position = include_position
        self._language = language
        self._speak_slowly = slow

        # page info
        self._text_to_speak_for_page: typing.Dict[int, str] = {}

    #
    # PRIVATE
    #

    def _end_page(self, page: Page):
        super(PDFToMP3, self)._end_page(page)
        self._text_to_speak_for_page[self._current_page_number] = "".join(
            [
                self._get_text_for_paragraph(p, i, self._current_page_number)
                for i, p in enumerate(
                    self._paragraphs_per_page[self._current_page_number]
                )
            ]
        )

    def _get_text_for_paragraph(
        self, paragraph: Paragraph, paragraph_number: int, page_number: int
    ):

        # text to speak
        text_to_speak_for_paragraph = ""

        # position
        if self._include_position:
            lbox: typing.Optional[Rectangle] = paragraph.get_previous_layout_box()
            assert lbox is not None
            text_to_speak_for_paragraph += "Page %d, paragraph %d, %s %s." % (
                page_number + 1,
                paragraph_number + 1,
                self._get_text_for_y(lbox),
                self._get_text_for_x(lbox),
            )
        # text of paragraph
        text_to_speak_for_paragraph += paragraph._text

        # force period if needed
        if text_to_speak_for_paragraph[-1] not in ["?", "!", "."]:
            text_to_speak_for_paragraph += ". "

        # return
        return text_to_speak_for_paragraph

    def _get_text_for_x(self, bounding_box: Rectangle) -> str:
        assert self._current_page
        w = self._current_page.get_page_info().get_width()
        assert w is not None
        xs: typing.List[Decimal] = [Decimal(0), w * Decimal(0.33), w * Decimal(0.66), w]
        x = bounding_box.x + bounding_box.width * Decimal(0.5)
        if xs[0] <= x <= xs[1]:
            return "left"
        if xs[1] <= x <= xs[2]:
            return "center"
        if xs[2] <= x <= xs[3]:
            return "right"
        return ""

    def _get_text_for_y(self, bounding_box: Rectangle) -> str:
        assert self._current_page
        h = self._current_page.get_page_info().get_height()
        assert h is not None
        assert h is not None
        ys: typing.List[Decimal] = [h, h * Decimal(0.66), h * Decimal(0.33), Decimal(0)]
        y = bounding_box.y
        if ys[1] <= y <= ys[0]:
            return "top"
        if ys[2] <= y <= ys[1]:
            return "middle"
        if ys[3] <= y <= ys[2]:
            return "bottom"
        return ""

    #
    # PUBLIC
    #

    @staticmethod
    def convert_pdf_to_mp3(pdf: Document) -> typing.Dict[int, bytes]:
        """
        This function converts a PDF to an MP3 file, returning its Path
        """
        sound_bytes_of_each_page: typing.Dict[int, bytes] = {}
        number_of_pages: int = int(pdf.get_document_info().get_number_of_pages() or 0)
        for page_nr in range(0, number_of_pages):
            # get Page object
            page: Page = pdf.get_page(page_nr)
            page_source: io.BytesIO = io.BytesIO(page["Contents"]["DecodedBytes"])

            # register EventListener
            cse: "PDFToMP3" = PDFToMP3()

            # process Page
            cse._event_occurred(BeginPageEvent(page))
            CanvasStreamProcessor(page, Canvas(), []).read(page_source, [cse])
            cse._event_occurred(EndPageEvent(page))

            # set in page
            sound_bytes_of_each_page[page_nr] = cse.convert_to_mp3()[0]

        # return
        return sound_bytes_of_each_page

    def convert_to_mp3(self) -> typing.Dict[int, bytes]:
        """
        This function creates and then returns the audio-file for the text spoken at the given page
        """
        sound_bytes_per_page: typing.Dict[int, bytes] = {}
        number_of_pages: int = len(self._text_to_speak_for_page.keys())
        for page_nr in range(0, number_of_pages):
            sound_for_page = gTTS(
                text=self._text_to_speak_for_page[page_nr], lang=self._language
            )

            # store in temporary location
            tmp_path: Path = Path(tempfile.NamedTemporaryFile().name)
            sound_for_page.save(tmp_path)

            # read bytes
            sound_bytes: typing.Optional[bytes] = None
            with open(tmp_path, "rb") as tmp_file_handle:
                sound_bytes = tmp_file_handle.read()
            assert sound_bytes is not None
            try:
                if tmp_path.exists():
                    tmp_path.unlink()
            except:
                pass
            # insert into dict
            sound_bytes_per_page[page_nr] = sound_bytes

        # return
        return sound_bytes_per_page



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This module contains the build information for this project
"""
from borb.license.version import Version
import setuptools


# open readme
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# requirements
required = [
    "cryptography>=37.0.4",     # license handling
    "fonttools>=4.22.1",        # TTF fonts
    "lxml>=4.9.1",              # XML processing
    "Pillow>=7.1.0",            # image processing
    "python-barcode>=0.13.1",   # generating barcodes
    "qrcode[pil]>=6.1",         # generating QR codes
    "requests>=2.24.0",         # generating images from a URL
    "setuptools>=51.1.1",       # standard
]

setuptools.setup(
    name="borb",
    version=Version.get_version(),
    author=Version.get_author(),
    author_email="joris.schellekens.1989@gmail.com",
    description="borb is a library for reading, creating and manipulating PDF files in python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jorisschellekens/borb",
    packages=setuptools.find_packages(include=["borb", "borb.*"]),
    include_package_data=True,
    install_requires=required,
    python_requires=">=3.6",
)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains everything needed to represent a low-level token in mathematical syntax
"""
import enum
import typing


class TokenType(enum.Enum):
    """
    This enum represents the various kinds of Token objects the Equation parser can encounter
    """

    COMMA = 1
    FUNCTION = 2
    LEFT_PARENTHESIS = 3
    NUMBER = 4
    OPERATOR = 5
    RIGHT_PARENTHESIS = 6
    VARIABLE = 7


class Token:
    """
    This module contains everything needed to represent a low-level token in mathematical syntax
    """

    def __init__(
        self,
        is_left_associative: typing.Optional[bool] = None,
        number_of_arguments: typing.Optional[int] = None,
        precedence: typing.Optional[int] = None,
        text: str = "",
        type: TokenType = TokenType.NUMBER,
    ):
        self._children: typing.List["Token"] = []
        self._is_left_associative: typing.Optional[bool] = is_left_associative
        self._number_of_arguments: typing.Optional[int] = number_of_arguments
        self._precedence: typing.Optional[int] = precedence
        self._text: str = text
        self._type: TokenType = type

    def get_children(self) -> typing.List["Token"]:
        """
        This function returns the children of this Token
        :return:    the children of this Token
        """
        return self._children

    def get_is_left_associative(self) -> typing.Optional[bool]:
        """
        This function return True if this Token represents a left-associative operator,
        False if it represents a right-associative operator, and None if this Token does
        not represent an operator
        :return:    True if this Token represents a left-associative operator
        """
        return self._is_left_associative

    def get_number_of_arguments(self) -> typing.Optional[int]:
        """
        This function returns the number of arguments the operator (represented by this Token)
        accepts, and None if this Token does not represent an operator
        :return:    the number of arguments this operator accepts
        """
        return self._number_of_arguments

    def get_precedence(self) -> typing.Optional[int]:
        """
        This function returns the precedence of the operator (represented by this Token)
        and None if this Token does not represent an operator
        :return:    the precedence of this operator
        """
        return self._precedence

    def get_text(self) -> str:
        """
        This function returns the text associated with this Token
        :return:    the text associated with this Token
        """
        return self._text

    def get_type(self) -> TokenType:
        """
        This function returns the TokenType of this Token
        :return:    the TokenType of this Token
        """
        return self._type



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This object represents the base class for everything related to IO in borb.
It has some convenience methods that allow you to specify how the object
should be persisted, as well as some methods to traverse the object-graph.
"""
import copy
import typing
from types import MethodType

import PIL


class PDFObject:
    """
    This object represents the base class for everything related to IO in borb.
    It has some convenience methods that allow you to specify how the object
    should be persisted, as well as some methods to traverse the object-graph.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(self):
        self._parent: typing.Optional["PDFObject"] = None
        self._is_inline: bool = False
        self._is_unique: bool = False
        self._reference: typing.Optional["Reference"] = None

    #
    # PRIVATE
    #

    @staticmethod
    def _to_json(self, memo_dict={}) -> typing.Any:

        # bool
        if isinstance(self, bool):
            return self

        # Boolean
        from borb.io.read.types import Boolean

        if isinstance(self, Boolean):
            return bool(self)

        # CanvasOperatorName
        from borb.io.read.types import CanvasOperatorName

        if isinstance(self, CanvasOperatorName):
            return str(self)

        # (borb) Decimal
        from borb.io.read.types import Decimal

        if isinstance(self, Decimal):
            return float(self)

        # (decimal) Decimal
        from decimal import Decimal as oDecimal

        if isinstance(self, oDecimal):
            return float(self)

        # float, int
        if isinstance(self, float) or isinstance(self, int):
            return self

        # bytes
        if isinstance(self, bytes):
            return str(self)

        # Dictionary
        from borb.io.read.types import Dictionary

        if isinstance(self, Dictionary):
            out: typing.Dict[str, typing.Any] = {}
            memo_dict[id(self)] = out
            for k, v in self.items():
                out[str(k)] = PDFObject._to_json(v, memo_dict)
            return out

        # dict
        if isinstance(self, dict):
            dict_out: typing.Dict[str, typing.Any] = {}
            memo_dict[id(self)] = dict_out
            for k, v in self.items():
                dict_out[str(k)] = PDFObject._to_json(v, memo_dict)
            return dict_out

        # Element
        from borb.io.read.types import Element

        if isinstance(self, Element):
            from borb.io.read.types import ET

            return str(ET.tostring(self))

        # Name
        from borb.io.read.types import Name

        if isinstance(self, Name):
            return str(self)

        # Stream
        # DUPLICATE: Dictionary

        # Function
        # DUPLICATE: Dictionary

        # String
        from borb.io.read.types import String

        if isinstance(self, String):
            return str(self)

        # HexadecimalString
        # DUPLICATE: String

        # List
        from borb.io.read.types import List

        if isinstance(self, List):
            list_out: typing.List[typing.Any] = []
            memo_dict[id(self)] = list_out
            for v in self:
                list_out += [PDFObject._to_json(v, memo_dict)]
            return list_out

        # Reference
        from borb.io.read.types import Reference

        if isinstance(self, Reference):
            return "%d %d R" % (self.generation_number or 0, self.object_number or 0)

        # PIL.Image.Image
        # TODO

        # default
        return None

    #
    # PUBLIC
    #

    @staticmethod
    def add_pdf_object_methods(non_borb_object: typing.Any) -> typing.Any:
        """
        This method allows you to pretend an object is actually a PDFObject.
        It adds all the methods that are present for a PDFObject.
        It also adds a utility hashing method for images (since PIL normally does not hash images)
        :param non_borb_object:
        :return:
        """

        def _deepcopy_and_add_methods(self, memodict={}):
            prev_function_ptr = self.__deepcopy__
            self.__deepcopy__ = None
            out = copy.deepcopy(self, memodict)
            self.__deepcopy__ = prev_function_ptr
            PDFObject.add_pdf_object_methods(out)
            return out

        def _get_parent(self) -> typing.Optional[PDFObject]:
            if "_parent" not in vars(self):
                setattr(self, "_parent", None)
            return self._parent

        def _get_reference(self) -> typing.Optional["Reference"]:  # type: ignore[name-defined]
            if "_reference" not in vars(self):
                setattr(self, "_reference", None)
            return self._reference

        def _get_root(self) -> PDFObject:
            p = self
            while p.get_parent() is not None:
                p = p.get_parent()
            return p

        def _is_inline(self) -> bool:
            if "_is_inline" not in vars(self):
                setattr(self, "_is_inline", False)
            return self._is_inline

        def _is_unique(self) -> bool:
            if "_is_unique" not in vars(self):
                setattr(self, "_is_unique", False)
            return self._is_unique

        def _pil_image_hash(self):
            w = self.width
            h = self.height
            pixels = [
                self.getpixel((0, 0)),
                self.getpixel((0, h - 1)),
                self.getpixel((w - 1, 0)),
                self.getpixel((w - 1, h - 1)),
            ]
            hashcode = 1
            for p in pixels:
                if isinstance(p, typing.List) or isinstance(p, typing.Tuple):
                    hashcode += 32 * hashcode + sum(p)
                else:
                    hashcode += 32 * hashcode + p
            return hashcode

        def _set_is_inline(self, is_inline: bool) -> PDFObject:
            if "_is_inline" not in vars(self):
                setattr(self, "_is_inline", False)
            self._is_inline = is_inline
            return self

        def _set_is_unique(self, is_unique: bool) -> PDFObject:
            if "_is_unique" not in vars(self):
                setattr(self, "_is_unique", False)
            self._is_unique = is_unique
            return self

        def _set_parent(self, parent: PDFObject) -> PDFObject:
            if "_parent" not in vars(self):
                setattr(self, "_parent", None)
            self._parent = parent
            return self

        def _set_reference(self, reference: "Reference") -> PDFObject:  # type: ignore[name-defined]
            if "_reference" not in vars(self):
                setattr(self, "_reference", None)
            self._reference = reference
            return self

        # inject these methods in the object
        non_borb_object.set_parent = MethodType(_set_parent, non_borb_object)
        non_borb_object.get_parent = MethodType(_get_parent, non_borb_object)
        non_borb_object.get_root = MethodType(_get_root, non_borb_object)
        non_borb_object.set_reference = MethodType(_set_reference, non_borb_object)
        non_borb_object.get_reference = MethodType(_get_reference, non_borb_object)
        non_borb_object.set_is_inline = MethodType(_set_is_inline, non_borb_object)
        non_borb_object.is_inline = MethodType(_is_inline, non_borb_object)
        non_borb_object.set_is_unique = MethodType(_set_is_unique, non_borb_object)
        non_borb_object.is_unique = MethodType(_is_unique, non_borb_object)
        non_borb_object.__deepcopy__ = MethodType(
            _deepcopy_and_add_methods, non_borb_object
        )

        # add a __hash__ method for PIL.Image.Image
        if isinstance(non_borb_object, PIL.Image.Image):
            non_borb_object.__hash__ = MethodType(_pil_image_hash, non_borb_object)  # type: ignore [assignment]

        # return
        return non_borb_object

    def get_parent(self) -> typing.Optional["PDFObject"]:
        """
        This function returns the parent of this PDFObject, or None if no such PDFObject exists
        :return:    the parent of this PDFObject
        """
        return self._parent

    def get_reference(self) -> typing.Optional["Reference"]:  # type: ignore[name-defined]
        """
        This function gets the Reference being used for this PDFObject
        :return:    the Reference being used for this PDFObject
        """
        return self._reference

    def get_root(self) -> typing.Optional["PDFObject"]:
        """
        This function returns the root (of the parent hierarchy) of this PDFObject, or None if no such PDFObject exists
        :return:    the root of this PDFObject
        """
        p: typing.Optional["PDFObject"] = self
        while p is not None and p.get_parent() is not None:
            p = p.get_parent()
        return p

    def is_inline(self) -> bool:
        """
        This function returns True if this PDFObject should be persisted inline, False otherwise
        :return:    whether this PDFObject should be persisted online
        """
        return self._is_inline

    def is_unique(self) -> bool:
        """
        This function returns True if this PDFObject should always be treated
        as if it is unique (for IO purposes), regardless of hashing equality.
        :return:    whether this PDFObject is unique
        """
        return self._is_unique

    def set_is_inline(self, is_inline: bool) -> "PDFObject":
        """
        This function sets the is_inline flag of this PDFObject.
        An inline object is always persisted immediately when needed, it is never turned into a reference.
        :param is_inline:   whether this PDFObject should be persisted inline, or not
        :return:            self
        """
        self._is_inline = is_inline
        return self

    def set_is_unique(self, is_unique: bool) -> "PDFObject":
        """
        This function sets the is_unique flag of this PDFObject.
        A unique object is always persisted as itself,
        or its own reference even if it should be equal to another PDFObject.
        :param is_unique:   whether this PDFObject should be unique or not
        :return:            self
        """
        self._is_unique = is_unique
        return self

    def set_parent(self, parent: "PDFObject") -> "PDFObject":
        """
        This function sets the parent (PDFObject) of this PDFObject
        :param parent:  the parent (PDFObject)
        :return:        self
        """
        self._parent = parent
        return self

    def set_reference(self, reference: "Reference") -> "PDFObject":  # type: ignore[name-defined]
        """
        This function sets the Reference to be used for this PDFObject
        :param reference:   the Reference to be used for this PDFObject
        :return:            self
        """
        self._reference = reference
        return self

    def to_json(self) -> typing.Any:
        """
        This function converts this PDFObject into a set of nested dictionaries, lists and primitives
        :return:    a JSON-like object
        """
        return PDFObject._to_json(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This class implements a (single word) find/replace functionality for PDF documents.
This class uses RegularExpressionTextExtraction to find the text,
RedactAnnotation to remove the text, and Paragraph to re-insert replacement text.
"""

import typing
from decimal import Decimal

from borb.pdf.document.document import Document
from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.canvas.layout.layout_element import Alignment
from borb.pdf.page.page import Page
from borb.pdf.canvas.color.color import Color
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.simple_font.font_type_1 import StandardType1Font
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.redact_annotation import RedactAnnotation


class SimpleFindReplace:
    """
    This class implements a (single word) find/replace functionality for PDF documents.
    This class uses RegularExpressionTextExtraction to find the text,
    RedactAnnotation to remove the text, and Paragraph to re-insert replacement text.
    """

    @staticmethod
    def sub(
        pattern: str,
        repl: str,
        doc: Document,
        page_range: typing.Optional[typing.List[int]] = None,
        repl_font: Font = StandardType1Font("Helvetica"),
        repl_font_color: typing.Optional[Color] = None,
        repl_font_horizontal_alignment: Alignment = Alignment.CENTERED,
        repl_font_size: typing.Optional[Decimal] = None,
    ) -> Document:
        """
        This function finds and replaces a regular expression in a PDF by a given piece of text
        :param pattern:                         the regular expression to be replaced
        :param repl:                            the replacement string
        :param doc:                             the PDF Document
        :param page_range:                      the pages on which to apply find/replace (if empty, this will default to all pages)
        :param repl_font:                       the Font to be used to insert the replacement text
        :param repl_font_color:                 the font_color to be used to isnert the replacement text (if empty, this will default to the original font_color)
        :param repl_font_size:                  the font_size to be used to insert the replacement text (if empty, this will default to the original font_size)
        :param repl_font_horizontal_alignment   the horizontal_alignment of the replacement text (defaults to Alignment.CENTERED)
        :return:                                the PDF Document
        """
        from borb.toolkit import RegularExpressionTextExtraction, PDFMatch

        number_of_pages: int = int(
            doc.get_document_info().get_number_of_pages() or Decimal(0)
        )
        matches_per_page: typing.Dict[
            int, typing.List[PDFMatch]
        ] = RegularExpressionTextExtraction.get_matches_for_pdf(pattern, doc)
        if page_range is None:
            page_range = [x for x in range(0, number_of_pages)]
        else:
            page_range = [x for x in page_range if x >= 0 and x < number_of_pages]
        for page_nr in page_range:

            # insert redaction annotations
            page: Page = doc.get_page(page_nr)
            for pdf_match in matches_per_page[page_nr]:
                for bounding_box in pdf_match.get_bounding_boxes():
                    page.add_annotation(RedactAnnotation(bounding_box.grow(Decimal(1))))

            # store original redact annotations
            # TODO

            # apply redact annotations
            page.apply_redact_annotations()

            # restore original redact annotations
            # TODO

            # insert new text (helvetica)
            for pdf_match in matches_per_page[page_nr]:

                # bounding box
                # fmt: off
                bb_x: Decimal = pdf_match.get_bounding_boxes()[0].get_x()
                bb_y: Decimal = pdf_match.get_bounding_boxes()[0].get_y()
                bb_h: Decimal = max([x.get_height() for x in pdf_match.get_bounding_boxes() if x.get_y() == bb_y])
                bb_w: Decimal = sum([x.get_width() for x in pdf_match.get_bounding_boxes() if x.get_y() == bb_y]) + Decimal(2)
                # fmt: on

                # determine font_size
                if repl_font_size is None:
                    repl_font_size = pdf_match.get_font_size()

                # determine font_color
                if repl_font_color is None:
                    repl_font_color = pdf_match.get_font_color()

                # put Paragraph
                if repl != "":
                    Paragraph(
                        repl,
                        font=repl_font,
                        font_size=repl_font_size,
                        font_color=repl_font_color,
                        horizontal_alignment=repl_font_horizontal_alignment,
                    ).paint(page, Rectangle(bb_x, bb_y + Decimal(1), bb_w, bb_h))

        # return
        return doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Farrow & Ball is a British manufacturer of paints and wallpapers largely based upon historic colour palettes and archives.
The company is particularly well known for the unusual names of its products.
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HexColor, RGBColor


class FarrowAndBall(HexColor):
    """
    Farrow & Ball is a British manufacturer of paints and wallpapers largely based upon historic colour palettes and archives.
    The company is particularly well known for the unusual names of its products.
    """

    COLOR_DEFINITION = {
        "all-white": "#faf9ef",
        "pointing": "#f9f3e3",
        "james-white": "#ece9d7",
        "white-tie": "#f2ead6",
        "clunch": "#e5dcc8",
        "blackened": "#dfdfdb",
        "tallow": "#f5ead2",
        "dimity": "#ece2d2",
        "great-white": "#e6e1dc",
        "strong-white": "#e9e7dd",
        "slipper-satin": "#eae4d3",
        "shaded-white": "#dad4c2",
        "new-white": "#f4e9d0",
        "lime-white": "#e9dfc9",
        "ringworld-ground": "#efe0c5",
        "matchstick": "#e6d8be",
        "house-white": "#f2e7c9",
        "off-white": "#e0d6c0",
        "string": "#dcceb0",
        "savage-ground": "#d7c5a9",
        "cord": "#d6c39f",
        "cream": "#e3cea7",
        "hay": "#d8c194",
        "light-stone": "#d1c1a0",
        "joas-white": "#e2d4be",
        "archive": "#decbb3",
        "stony-ground": "#cdc2ad",
        "cornforth-white": "#d4cfc7",
        "elephants-breath": "#cdc3b7",
        "lamp-room-gray": "#b7b6ae",
        "light-blue": "#b9beb5",
        "hardwick-white": "#b6b1a1",
        "blue-gray": "#b5b7a7",
        "french-gray": "#b4b19a",
        "pigeon": "#9fa195",
        "card-room-green": "#899081",
        "old-white": "#cdc3ad",
        "bone": "#cdc4ae",
        "fawn": "#ccbfa5",
        "biscuit": "#bca783",
        "light-gray": "#b2a894",
        "mouses-back": "#968975",
        "smoked-trout": "#bea895",
        "dead-salmon": "#b09b8c",
        "london-stone": "#b2a18d",
        "buff": "#b59e82",
        "drab": "#9b8a6f",
        "dauphin": "#928069",
        "pink-ground": "#ecd8ca",
        "calamine": "#e8d6d1",
        "setting-plaster": "#dbc2b0",
        "dutch-pink": "#d5b08a",
        "entrance-hall-pink": "#c39e80",
        "flower-pink": "#d6a787",
        "pintment-pink": "#c7997e",
        "red-earth": "#bb7b69",
        "porphyry-pink": "#b17b6c",
        "fox-red": "#a8725d",
        "loggia": "#aa6d5b",
        "blazer": "#ac514c",
        "book-room-red": "#a3695b",
        "picture-gallery-red": "#975b4f",
        "rectory-red": "#99434c",
        "radiocchio": "#8f4a51",
        "etruscan-red": "#7e534a",
        "eating-room-red": "#864e4f",
        "pale-hound": "#e6ddb8",
        "hound-lemon": "#e2d2a4",
        "farrows-cream": "#e8d4b0",
        "dayroom-yellow": "#f2dc9e",
        "dorset-cream": "#eacfa2",
        "gervase-yellow": "#dcc98f",
        "yellow-ground": "#efca89",
        "citron": "#eecc81",
        "ciara-yellow": "#e1c17e",
        "straw": "#d7b37e",
        "print-room-yellow": "#dbb678",
        "babouche": "#eabe67",
        "sudbury-yellow": "#d9b475",
        "octagon-yellow": "#d1ae72",
        "orangery": "#e0ae6c",
        "india-yellow": "#c09759",
        "cane": "#caa26c",
        "sand": "#ba9166",
        "green-ground": "#dbd8b6",
        "cooking-apple-green": "#c4c3a2",
        "vert-de-terre": "#babca4",
        "ball-green": "#bab393",
        "stone-white": "#b3ac8c",
        "green-stone": "#a59c78",
        "lichen": "#9e9f87",
        "olive": "#8b896b",
        "sutcliffe-green": "#899677",
        "pea-green": "#849779",
        "calke-green": "#77866a",
        "minster-green": "#586552",
        "powder-blue": "#acbdb4",
        "chappell-green": "#93a595",
        "saxon-green": "#99a680",
        "folly-green": "#95aa86",
        "breakfast-room-green": "#94a588",
        "arsenic": "#88b49b",
        "pale-powder": "#dde1d5",
        "borrowed-light": "#d8dfdc",
        "skylight": "#c6ccc7",
        "teresas-green": "#c2cfc2",
        "green-blue": "#aebdb1",
        "castle-gray": "#8d988c",
        "blue-ground": "#a7c9c8",
        "dix-blue": "#9ab1ab",
        "ballroom-blue": "#8faba7",
        "sugar-bag-light": "#81a09a",
        "oval-room-blue": "#8b9d9b",
        "berrington-blue": "#7e9495",
        "parma-gray": "#b1c0c2",
        "lulworth-blue": "#a2b8c7",
        "stone-blue": "#77989f",
        "chinese-blue": "#6f96a6",
        "cooks-blue": "#6f94b3",
        "pitch-blue": "#61718e",
        "wainscot": "#7e634e",
        "down-pipe": "#626665",
        "hague-blue": "#3e4e56",
        "brinjal": "#5a4348",
        "mahogany": "#524947",
        "off-black": "#424548",
        "green-smoke": "#717b70",
        "monkey-puzzle": "#4b5853",
        "caggiage-green": "#434c47",
        "railings": "#43474a",
        "studio-green": "#474d4a",
        "black-blue": "#3d4345",
    }

    #
    # CONSTRUCTOR
    #

    def __init__(self, color_name: str):
        assert color_name in FarrowAndBall.COLOR_DEFINITION
        self.color_name: str = color_name
        super(FarrowAndBall, self).__init__(FarrowAndBall.COLOR_DEFINITION[color_name])

    #
    # PRIVATE
    #

    def __deepcopy__(self, memodict={}):
        return FarrowAndBall(self.color_name)

    #
    # PUBLIC
    #

    @staticmethod
    def find_nearest_farrow_and_ball_color(color: Color) -> "FarrowAndBall":
        """
        This function find the nearest `Farrow and Ball` equivalent for a given Color
        """
        rgb_color_001: RGBColor = color.to_rgb()
        min_dist: typing.Optional[Decimal] = None
        color_with_min_dist: typing.Optional[str] = None
        for n, c in FarrowAndBall.COLOR_DEFINITION.items():
            rgb_color_002: RGBColor = HexColor(c)
            d: Decimal = (
                (rgb_color_001.red - rgb_color_002.red) ** 2
                + (rgb_color_001.green - rgb_color_002.green) ** 2
                + (rgb_color_001.blue - rgb_color_002.blue) ** 2
            )
            if min_dist is None or d < min_dist:
                min_dist = d
                color_with_min_dist = n
        assert min_dist is not None
        assert color_with_min_dist is not None
        return FarrowAndBall(color_with_min_dist)

    def get_name(self) -> str:
        """
        This function returns the name of this `Farrow and Ball` color
        """
        return self.color_name
